// https://github.com/dlang/dmd/issues/23431
// test linking against debug build of druntime.lib producing
// druntime.lib(package.obj) : warning LNK4255: library contain multiple objects of the same name; linking object as if no debug info

void main()
{
}
